/*
 * Copyright 2005,2006 Red Hat, Inc.
 *
 * Author: Chris Feist <cfeist@redhat.com>
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <unistd.h>
#include <netdb.h>
#include <errno.h>
#include <ldap.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rpcsvc/ypclnt.h>
#include <rpc/types.h>
#include "automount.h"

#define MODPREFIX "nsswitch: "

#define MAPTYPE_FILE 1
#define MAPTYPE_PROGRAM 2

int underscore_to_dot = 0;

static int isfilemap(const char *);
static int isypmap(const char *);
static int isldapmap(const char *);

char *make_mapname(const char *loc, const char *type)
{
	int retsize, needs_etc = 0;
	char *retval;

	retsize = strlen(loc) + strlen(type) + 2;
       	if (*loc != '/') {
		retsize += 5; /* " "/etc/" */
		needs_etc = 1;
	}

	retval = malloc(retsize);
	if (!retval)
		return NULL;

	snprintf(retval, retsize, "%s:%s%s", type,
		 needs_etc ? "/etc/" : "", loc);

	return retval;
}

/*
 * Function which takes in a partial map name (ie. auto.misc), parses the
 * nsswitch.conf file and returns a valid map name (ie. yp:auto.misc or
 * file:/etc/auto.misc).  "cur" is the current map we are reading.  This
 * is used when determining the source of an included map.  For example.
 * if we are parsing /etc/auto.master, and it has the line "+auto.master",
 * we need to grab the auto.master from nis, if nis is specified in
 * nsswitch.conf.
 */
char *get_nsswitch_map(const char *loc, char *cur)
{
	char buf[1024];
	char *ordering;
	const char *automount_str = "automount:";
	char *comment = NULL;
	FILE *nsswitch;
	int found_automount = 0;
	char *retval = NULL;
	int retsize = 0;
	char *ypmapname = (char *)loc;

	debug(MODPREFIX "called nsswitch with: '%s' '%s'", loc, cur);
	nsswitch = fopen(_PATH_NSSWITCH_CONF, "r");
	if (!nsswitch) {
		error(MODPREFIX "Unable to open %s", _PATH_NSSWITCH_CONF);
		return NULL;
	}

	while ((ordering = fgets((char *)buf, sizeof(buf), nsswitch))) {
		if ((comment = strchr(ordering,'#')))
			*comment = '\0';
		while (isspace(*ordering)) ordering++;
		if (!strncmp(ordering, automount_str, sizeof(automount_str))) {
			ordering += strlen(automount_str);
			found_automount = 1;
			break;
		}
	}

	fclose(nsswitch);

	if (!found_automount)
		return NULL;

	/*
	 *  We support an option to convert underscores to dots when looking
	 *  up yp maps.
	 */
	if (underscore_to_dot) {
		char *u;

		ypmapname = strdup(loc);
		if (!ypmapname)
			return NULL;
		while ((u = strchr(ypmapname, '_')))
			*u = '.';
	}

	while (*ordering != '\0') {
		while (isspace(*ordering) && (*ordering != '\0')) ordering++;
		if (!strncmp(ordering, "files", 5)) {
			switch (isfilemap(loc)) {
			case MAPTYPE_FILE:
				retval = make_mapname(loc, "file");
				debug("comparing %s to %s\n", retval+5,cur);
				if (cur && !strcmp(cur, retval+5)) {
					free(retval);
					retval = NULL;
					break;
				}
				return retval;
			case MAPTYPE_PROGRAM:
				retval = make_mapname(loc, "program");
				debug("comparing %s to %s\n", retval+5,cur);
				if (cur && !strcmp(cur, retval+5)) {
					free(retval);
					retval = NULL;
					break;
				}
				return retval;
			default: // filemap doesn't exist
				break;
			}

		} else if ((!strncmp(ordering, "yp", 2) ||
			    !strncmp(ordering,"nis", 3)) &&
			   isypmap(ypmapname)) {
			retsize = strlen(ypmapname) + 4;
			retval = malloc(retsize);
			snprintf(retval, retsize, "yp:%s", ypmapname);
			return retval;
		} else if (!strncmp(ordering, "ldap", 4) && isldapmap(loc)) {
			retsize = strlen(loc) + 6;
			retval = malloc(retsize);
			snprintf(retval, retsize, "ldap:%s", loc);
			return retval;
		}
		while (!isspace(*ordering) && (*ordering != '\0')) ordering++;
	}
	error(MODPREFIX "couldn't find map %s", loc);
	return retval;
}

/*
 * Function takes in a filename and tests if it exists in "/etc/"
 * Returns: MAPTYPE_FILE if it is not executable, MAPTYPE_PROGRAM if it
 * is executable and 0 if it doesn't exists or has incorrect permissions.
 */

static int isfilemap(const char *loc)
{
	struct stat st;
	int ret = 0;	
	char *realfilemap;

	if (*loc != '/') {
		realfilemap = malloc(strlen(loc) + 6); /* '/etc/' + '\0' */
		if (!realfilemap) {
			crit(MODPREFIX "malloc failed.");
			return 0;
		}
		snprintf(realfilemap, strlen(loc) + 6, "/etc/%s", loc);
		ret = stat(realfilemap, &st);
		free(realfilemap);
	} else
		ret = stat(loc, &st);

	if (!ret) {
		if (st.st_uid != 0) {
			error(MODPREFIX "file /etc/%s exists but is not"
					" owned by root.", loc);
			return 0;
		} else if (st.st_mode & S_IRUSR) {
			if (st.st_mode & S_IXUSR) 
				return MAPTYPE_PROGRAM;
			else
				return MAPTYPE_FILE;
		}
	}
	debug(MODPREFIX "stat(%s) returned errno %d\n", loc, errno);
	return 0;
}

/*
 * Function takes in a yp map name and returns
 * 1 if it exists or 0 if it doesn't.
 *
 * Some of this code borrowed from ypcat
 */

static int isypmap(const char *loc)
{
	int err;
	char *domainname;
	char *hostname;
	unsigned int order;

	if ((err = yp_get_default_domain(&domainname)) != YPERR_SUCCESS) {
		error(MODPREFIX "unable to get default yp domain");
		return 0;
	}

	if ((err = yp_order(domainname, loc, &order)) != YPERR_SUCCESS) {
		/*
		 * Some nisplus servers, when running in nis compatibility
		 * mode, will not support yp_order.  So, we check for the
		 * existence of the map with the yp_master call.
		 */
		err = yp_master(domainname, loc, &hostname);
		if (err != YPERR_SUCCESS) {
			debug(MODPREFIX "unable to find map, %s in domain, %s",
			      loc, domainname);
			return 0;
		}
		free(hostname);
	}

	return 1;
}

/*
 *  boolean function: it returns 1 if 'mapname' exists in the ldap sever, and
 *  0 if it can't be found.
 */
static int ldap_map_exists(const char *mapname,
			   const char *class, const char *attr)
{
	LDAP *ld = NULL;
	LDAPControl *server = NULL, *client = NULL;
	LDAPMessage *messages = NULL, *entry = NULL;
	char *attrs[2];
	char filter[LINE_MAX] = "";
	int result, ret = 0;
	int c;

	/* Initialize the LDAP library, pointing it at the default server. */
	ld = ldap_init(NULL, LDAP_PORT);
	if (ld == NULL) {
		crit(MODPREFIX "error initializing LDAP\n");
		return 0;
	}

	/* Try to switch to protocol 3. */
	c = 3;
	if (ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &c)
	    != LDAP_SUCCESS) {
		/* Just retry with the default protocol version. */
		ldap_unbind(ld);
		ld = ldap_init(NULL, LDAP_PORT);
		if (ld == NULL) {
			crit(MODPREFIX "error initializing LDAP\n");
			return 0;
		}
	}

	/* Connect to the server anonymously. */
	result = ldap_simple_bind_s(ld, NULL, NULL);
	if (result != LDAP_SUCCESS) {
		crit(MODPREFIX "error binding to ldap server: %s\n",
		     ldap_err2string(result));
		goto out;
	}

	/* We only want the key and value attributes from entries which
	 * match the query we'll perform. */
	attrs[0] = LDAP_NO_ATTRS;
	attrs[1] = NULL;

	/* Set the filter to only find the map we're looking at. */
	snprintf(filter, sizeof(filter), "(&(objectclass=%s)(%s=%s))",
		 class, attr, mapname);

	debug(MODPREFIX "searching for map %s with filter %s\n",
	      mapname, filter);
	/* Perform a synchronous query to find the DN of the map. */
	result = ldap_search_ext_s(ld, NULL,
				   LDAP_SCOPE_SUBTREE,
				   filter,
				   attrs, FALSE,
				   &server, &client,
				   NULL,
				   LDAP_NO_LIMIT,
				   &messages);
	if (result != LDAP_SUCCESS)
		goto out;

	/* We expected only one match.  Pull it from the results list. */
	entry = ldap_first_entry(ld, messages);
	if (entry != NULL)
		ret = 1; /* success */

out:
	ldap_unbind(ld);
	if (server)
		ldap_control_free(server);
	if (client)
		ldap_control_free(client);
	if (messages)
		ldap_msgfree(messages);

	return ret;

}

static int isldapmap(const char *loc)
{
	/*
	 * search for the three schemas:
	 *   (&(objectclass=nisMap)(nisMapName=%s))
	 *   (&(objectclass=automountMap)(ou=%s))
	 *   (&(objectclass=automountMap)(automountMapName=%s))
	 */
	if (ldap_map_exists(loc, "nisMap", "nisMapName")) {
		debug(MODPREFIX "map %s found on the ldap server\n", loc);
		return 1;
	}
	if (ldap_map_exists(loc, "automountMap", "ou")) {
		debug(MODPREFIX "map %s found on the ldap server\n", loc);
		return 1;
	}
	if (ldap_map_exists(loc, "automountMap", "automountMapName")) {
		debug(MODPREFIX "map %s found on the ldap server\n", loc);
		return 1;
	}

	debug(MODPREFIX "map %s not found on the ldap server\n", loc);
	return 0;
}

#ifdef STANDALONE
int do_debug = 0;

void print_usage(const char *prog)
{
	fprintf(stderr, "Usage %s [-u] <mapname>\n", basename(prog));
}

int main(int argc, char **argv)
{
	char *map;
	int  opt;

	if (argc < 2) {
		print_usage(argv[0]);
		exit(1);
	}

	while ((opt = getopt(argc, argv, "u")) != -1) {
		switch (opt) {
		case 'u':
			underscore_to_dot = 1;
			break;
		case '?':
		default:
			print_usage(argv[0]);
			exit(1);

		}
	}

	map = get_nsswitch_map(argv[optind], (char *)"auto.master");
	if (!map)
		exit(2);

	fprintf(stdout, "%s\n", map);
	free(map);
	exit(0);
}
#endif
