#ident "$Id: mount_nfs.c,v 1.12 2004/05/18 12:20:08 raven Exp $"
/* ----------------------------------------------------------------------- *
 *   
 * mount_nfs.c - Module for Linux automountd to mount an NFS filesystem,
 *               with fallback to symlinking if the path is local
 *
 *   Copyright 1997 Transmeta Corporation - All Rights Reserved
 *   Copyright 1999-2000 Jeremy Fitzhardinge <jeremy@goop.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <linux/nfs.h>
#include <linux/nfs2.h>
#include <ctype.h>

#define MODULE_MOUNT
#include "automount.h"
#include "replicated.h"

#define MODPREFIX "mount(nfs): "

int mount_version = AUTOFS_MOUNT_VERSION;	/* Required by protocol */

static int udpproto;
static short port_discard;

static struct mount_mod *mount_bind = NULL;

int mount_init(void **context)
{
	struct protoent *udp;
	struct servent *port_dis;

	/* These are context independent */
	udp = getprotobyname("udp");
	udpproto = udp ? udp->p_proto : 0;
	port_dis = getservbyname("discard", "udp");

	if (port_dis)
		port_discard = port_dis->s_port;
	else
		port_discard = htons(9);	/* 9 is the standard discard port */

	/* Make sure we have the local mount method available */
	if (!mount_bind)
		mount_bind = open_mount("bind", MODPREFIX);

	return !mount_bind;
}

void seed_random(void)
{
	int fd, ret;
	unsigned int seed;

	if (!ap.random_multimount)
		return;

	fd = open("/dev/random", O_RDONLY);
	if (fd < 0) {
		srandom(time(NULL));
		return;
	}

	read(fd, &seed, sizeof(seed));
	srandom(seed);
	close(fd);
}

int mount_mount(const char *root, const char *name, int name_len,
		const char *what, const char *fstype, const char *options,
		void *context)
{
	char *fullpath, buf[MAX_ERR_BUF];
	struct host *this, *hosts = NULL;
	unsigned int save_ghost = ap.ghost;
	char *nfsoptions = NULL;
	unsigned int vers;
	int nosymlink = 0;
	int len, rlen, status, err, existed = 1;
	int ro = 0;

	debug(MODPREFIX " root=%s name=%s what=%s, fstype=%s, options=%s",
	      root, name, what, fstype, options);

	seed_random();

	/* Extract "nosymlink" pseudo-option which stops local filesystems
	   from being symlinked */
	if (options) {
		const char *comma;
		char *nfsp;
		int len = strlen(options) + 1;

		nfsp = nfsoptions = alloca(len + 1);
		if (!nfsoptions)
			return 1;

		memset(nfsoptions, '\0', len + 1);

		for (comma = options; *comma != '\0';) {
			const char *cp;
			const char *end;

			while (*comma == ',')
				comma++;

			/* Skip leading white space */
			while (*comma == ' ' || *comma == '\t')
				comma++;

			cp = comma;
			while (*comma != '\0' && *comma != ',')
				comma++;

			/* Skip trailing white space */
			end = comma - 1;
			while (*comma == ' ' || *comma == '\t')
				end--;

#if 0
			debug(MODPREFIX "*comma=%x %c  comma=%p %s cp=%p %s "
			      "nfsoptions=%p nfsp=%p end=%p used=%d len=%d\n",
			      *comma, *comma, comma, comma, cp, cp,
			      nfsoptions, nfsp, nfsoptions + len,
			      nfsp - nfsoptions, len);
#endif
			if (strncmp("nosymlink", cp, end - cp + 1) == 0)
				nosymlink = 1;
			else {
				/* and jump over trailing white space */
				memcpy(nfsp, cp, comma - cp + 1);
				nfsp += comma - cp + 1;
			}
		}

		debug(MODPREFIX "nfs options=\"%s\", nosymlink=%d",
		      nfsoptions, nosymlink);
	}

	if (strcmp(fstype, "nfs4") == 0)
		vers = NFS4_VERS_MASK | NFS_PROTO_MASK;
	else
		vers = NFS_VERS_MASK | NFS_PROTO_MASK;

	if (!parse_location(&hosts, what)) {
		warn(MODPREFIX "no hosts available");
		return 1;
	}
	prune_host_list(&hosts, vers, nfsoptions);

	if (!hosts) {
		warn(MODPREFIX "no hosts available");
		return 1;
	}

	/* Construct and perhaps create mount point directory */

	/* Root offset of multi-mount */
	if (*name == '/' && name_len == 1) {
		rlen = strlen(root);
		name_len = 0;
	/* Direct mount name is absolute path so don't use root */
	} else if (*name == '/')
		rlen = 0;
	else
		rlen = strlen(root);

	fullpath = alloca(rlen + name_len + 2);
	if (!fullpath) {
		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
		error(MODPREFIX "alloca: %s", estr);
		free_host_list(&hosts);
		return 1;
	}

	if (name_len) {
		if (rlen)
			len = sprintf(fullpath, "%s/%s", root, name);
		else
			len = sprintf(fullpath, "%s", name);
	} else
		len = sprintf(fullpath, "%s", root);
	fullpath[len] = '\0';

	debug(MODPREFIX "calling mkdir_path %s", fullpath);

	status = mkdir_path(fullpath, 0555);
	if (status && errno != EEXIST) {
		error(MODPREFIX "mkdir_path %s failed: %m", fullpath);
		return 1;
	}

	if (!status)
		existed = 0;

	/*
	 * We need to stop the bind mount module from removing the
	 * mount point directory if a bind attempt fails so abuse
	 * the ap.ghost field for this.
	 */
	ap.ghost = 1;

	this = hosts;
	while (this) {
		char *loc, *port_opt = NULL;

		if (is_mounted(fullpath)) {
			unlink(AUTOFS_LOCK);
			error(MODPREFIX
			      "warning: %s is already mounted", fullpath);
			free_host_list(&hosts);
			break;
		}

		/*
		 * If the "port" option is specified, then we don't want
		 * a bind mount. Use the "port" option if you want to
		 * avoid attempting a local bind mount, such as when
		 * tunneling NFS via localhost.
		 */
		if (nfsoptions && *nfsoptions)
			port_opt = strstr(nfsoptions, "port=");

		/* Port option specified, don't try to bind */
		if (!port_opt && this->proximity == PROXIMITY_LOCAL) {
			/* Local host -- do a "bind" */
			const char *bind_options = ro ? "ro" : "";

			debug(MODPREFIX "%s is local, attempt bind mount",
			      name);

			/* mount_bind will take the lock */
			err = mount_bind->mount_mount(root, name, name_len,
					       this->path, "bind", bind_options,
					       mount_bind->context);

			/* Success - we're done */
			if (!err) {
				free_host_list(&hosts);
				ap.ghost = save_ghost;
				return 0;
			}

			/* No hostname, can't be NFS */
			if (!this->name) {
				this = this->next;
				continue;
			}
		}

		/* Not a local host - do an NFS mount */

		loc = malloc(strlen(this->name) + 1 + strlen(this->path) + 1);
		strcpy(loc, this->name);
		strcat(loc, ":");
		strcat(loc, this->path);

		wait_for_lock();
		if (nfsoptions && *nfsoptions) {
			debug(MODPREFIX "calling mount -t %s " SLOPPY 
			      "-o %s %s %s", fstype, nfsoptions, loc, fullpath);
			err = spawn_mount(LOG_NOTICE, "-t", fstype,
					  SLOPPYOPT "-o",
					  nfsoptions, loc, fullpath, NULL);
		} else {
			debug(MODPREFIX "calling mount -t %s %s %s",
			      fstype, loc, fullpath);
			err = spawn_mount(LOG_NOTICE, "-t", fstype,
					  loc, fullpath, NULL);
		}
		unlink(AUTOFS_LOCK);

		if (!err) {
			info(MODPREFIX "mounted %s on %s", loc, fullpath);
			free(loc);
			free_host_list(&hosts);
			ap.ghost = save_ghost;
			return 0;
		}

		free(loc);
		this = this->next;
	}

	free_host_list(&hosts);
	ap.ghost = save_ghost;

	/* If we get here we've failed to complete the mount */
	if ((!ap.ghost && name_len) || !existed)
		rmdir_path(name);

	error(MODPREFIX "nfs: mount failure %s on %s", what, fullpath);
	return 1;
}

int mount_done(void *context)
{
	return mount_bind->mount_done(mount_bind->context);
}
